# frozen_string_literal: true

module Kraftwerk
  module Middleware
    class RequestTelemetry
      ROUTER_PARAMS = 'router.params'
      PATH = 'PATH_INFO'
      METHOD = 'REQUEST_METHOD'
      QUERY_STRING = 'QUERY_STRING'
      RACK_INPUT = 'rack.input'

      attr_reader :app, :kraftwerk_app, :telemetry

      def initialize(app, kraftwerk_app, telemetry: nil)
        @app = app
        @kraftwerk_app = kraftwerk_app
        @telemetry ||= kraftwerk_app::Components['telemetry']
      end

      def call(env)
        metadata = { app: kraftwerk_app, time: Time.now }
        start_data = request_start_data(env)
        telemetry.execute([:kraftwerk, :request, :start], start_data, metadata)

        start_time = Process.clock_gettime(Process::CLOCK_MONOTONIC)
        app.call(env).tap do |response|
          telemetry.execute(
            [:kraftwerk, :request, :finish],
            request_finish_data(start_data, response, start_time),
            metadata
          )
        end
      end

      private

      def request_start_data(env)
        {
          method: env[METHOD],
          path: env[PATH],
          params: request_params(env)
        }
      end

      def request_finish_data(start_data, response, start_time)
        duration = Process.clock_gettime(Process::CLOCK_MONOTONIC) - start_time
        code, _headers, body = response

        start_data.merge(
          {
            duration: duration,
            code: code,
            body: body[0]
          }
        )
      end

      # Combines params from query string and body into a single hash
      def request_params(env)
        params_from_query_string = env[QUERY_STRING] ? CGI.parse(env[QUERY_STRING]) : {}
        rack_input = env[RACK_INPUT].dup.read
        params_from_input = rack_input.length > 0 ? CGI.parse(rack_input) : {}
        params_from_router = env[ROUTER_PARAMS] || {}
        params_from_query_string.merge(params_from_input).merge(params_from_router)
      end
    end
  end
end
